@@ -1,5 +1,5 @@
 /*
-   Copyright (c) 2006, 2007, 2008 Z RESEARCH, Inc. <http://www.zresearch.com>
+   Copyright (c) 2006-2009 Z RESEARCH, Inc. <http://www.zresearch.com>
    This file is part of GlusterFS.
 
    GlusterFS is free software; you can redistribute it and/or modify
@@ -36,8 +36,8 @@ struct random_sched_struct {
 
 struct random_struct {
   int32_t child_count;
-  int32_t refresh_interval;
-  int64_t min_free_disk;
+  uint32_t refresh_interval;
+  uint32_t min_free_disk;
   struct timeval last_stat_entry;
   struct random_sched_struct *array;
   pthread_mutex_t random_mutex;
