@@ -1,5 +1,5 @@
 /*
-   Copyright (c) 2006, 2007, 2008 Z RESEARCH, Inc. <http://www.zresearch.com>
+   Copyright (c) 2006-2009 Z RESEARCH, Inc. <http://www.zresearch.com>
    This file is part of GlusterFS.
 
    GlusterFS is free software; you can redistribute it and/or modify
@@ -26,8 +26,8 @@
 #endif
 
 typedef struct {
-  int encrypt_write;
-  int decrypt_read;
+	gf_boolean_t encrypt_write;
+	gf_boolean_t decrypt_read;
 } rot_13_private_t;
 
 #endif /* __ROT_13_H__ */
