@@ -1,5 +1,5 @@
 /*
-   Copyright (c) 2007, 2008 Z RESEARCH, Inc. <http://www.zresearch.com>
+   Copyright (c) 2007-2009 Z RESEARCH, Inc. <http://www.zresearch.com>
    This file is part of GlusterFS.
 
    GlusterFS is free software; you can redistribute it and/or modify
@@ -28,8 +28,12 @@
 #include <stdlib.h>
 #include <fuse/fuse_lowlevel.h>
 
+#define GLUSTERFS_XATTR_LEN_MAX  65536
+
 uint64_t req_callid (fuse_req_t req);
 
+size_t fuse_dirent_size (size_t dname_len);
+
 int32_t
 fuse_reply_vec (fuse_req_t req,
 		struct iovec *vector,
